@@ -48,17 +48,11 @@ def parse_args():
 
 
 def main():
-    state = SlimtaState()
-    state.load_config(*parse_args())
+    argparser, args = parse_args()
+    state = SlimtaState(args)
+    state.load_config(argparser=argparser)
 
     state.loop()
 
 
-def worker_main():
-    state = SlimtaState()
-    state.load_config(*parse_args())
-
-    state.worker_loop()
-
-
 # vim:et:fdm=marker:sts=4:sw=4:ts=4
