@@ -25,12 +25,14 @@ import sys
 import os
 import os.path
 import warnings
+import logging
 from importlib import import_module
 from functools import wraps
 from contextlib import contextmanager
 
 from config import Config, ConfigError, ConfigInputStream
-from gevent import sleep
+from gevent import sleep, socket
+from gevent.event import AsyncResult
 import slimta.system
 
 from .validation import ConfigValidation, ConfigValidationError
@@ -42,14 +44,16 @@ class SlimtaState(object):
     _global_config_files = [os.path.expanduser('~/.slimta/slimta.conf'),
                             '/etc/slimta/slimta.conf']
 
-    def __init__(self):
+    def __init__(self, args):
         self.program = os.path.basename(sys.argv[0])
-        self.args = None
+        self.args = args
         self.cfg = None
+        self.loop_interrupt = AsyncResult()
+        self.cached_listeners = {}
+        self.listeners = {}
         self.edges = {}
         self.queues = {}
         self.relays = {}
-        self._celery = None
 
     @contextmanager
     def _with_chdir(self, new_dir):
@@ -62,15 +66,19 @@ class SlimtaState(object):
 
     @contextmanager
     def _with_sighandlers(self):
-        from signal import SIGTERM
+        from signal import SIGTERM, SIGHUP
         from gevent import signal
         def handle_term():
             sys.exit(0)
+        def handle_hup():
+            self.loop_interrupt.set('reload')
         old_term = signal(SIGTERM, handle_term)
+        old_hup = signal(SIGHUP, handle_hup)
         try:
             yield
         finally:
             signal(SIGTERM, old_term)
+            signal(SIGHUP, old_hup)
 
     def _try_configs(self, files):
         for config_file in files:
@@ -83,26 +91,30 @@ class SlimtaState(object):
                         return Config(config_base), config_file
         return None, None
 
-    def load_config(self, argparser, args):
-        if self.args:
-            return
-        self.args = args
-
-        if args.process_name:
-            self.program = args.process_name
+    def load_config(self, argparser=None):
+        if self.args.process_name:
+            self.program = self.args.process_name
 
         files = self._global_config_files
-        if args.config:
-            files = [args.config]
+        if self.args.config:
+            files = [self.args.config]
 
         self.cfg, config_file = self._try_configs(files)
+        err = None
         if self.cfg:
             try:
                 ConfigValidation.check(self.cfg, self.program)
             except ConfigValidationError as e:
-                argparser.error(str(e))
+                err = str(e)
         else:
-            argparser.error('No configuration files found!')
+            err = 'No configuration files found!'
+
+        if err:
+            if argparser:
+                argparser.error(err)
+            else:
+                logging.getLogger('slimta.app').error(err)
+                sys.exit(2)
 
     def drop_privileges(self):
         process_options = self.cfg.process.get(self.program)
@@ -179,6 +191,51 @@ class SlimtaState(object):
         factory = self._import_symbol(options.factory)
         return factory(options, *extra)
 
+    def _copy_listener(self, listener):
+        if isinstance(listener, socket.socket):
+            fd = listener.fileno()
+            family = listener.family
+            type = listener.type
+            proto = listener.proto
+            return socket.fromfd(fd, family, type, proto)
+        return listener
+
+    def _get_listener(self, options, defaults):
+        key = hash(tuple(options.iteritems()))
+        if key in self.cached_listeners:
+            existing = self.cached_listeners[key]
+            listener_copy = self._copy_listener(existing)
+            self.listeners[key] = listener_copy
+            return listener_copy
+        type = options.get('type', 'tcp')
+        new_listener = None
+        if type in ('tcp', 'udp', 'unix'):
+            if type == 'tcp':
+                interface = options.get('interface', defaults.get('interface'))
+                port = int(options.get('port', defaults.get('port')))
+                new_listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                address = (interface, port)
+            elif type == 'udp':
+                interface = options.get('interface', defaults.get('interface'))
+                port = int(options.get('port', defaults.get('port')))
+                new_listener = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+                address = (interface, port)
+            elif type == 'unix':
+                new_listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+                address = options.get('path', defaults.get('path'))
+            new_listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+            new_listener.setblocking(0)
+            new_listener.bind(address)
+            if type != 'udp':
+                backlog = int(options.get('backlog', 256))
+                new_listener.listen(backlog)
+        elif type == 'custom':
+            new_listener = self._load_from_custom(options)
+        else:
+            raise ValueError('Unknown listener type: {0}'.format(type))
+        self.listeners[key] = new_listener
+        return new_listener
+
     def _start_relay(self, name, options=None):
         if name in self.relays:
             return self.relays[name]
@@ -293,10 +350,6 @@ class SlimtaState(object):
         elif options.type == 'proxy':
             from slimta.queue.proxy import ProxyQueue
             new_queue = ProxyQueue(relay)
-        elif options.type == 'celery':
-            from slimta.celeryqueue import CeleryQueue
-            backoff = build_backoff_function(options.get('retry'))
-            new_queue = CeleryQueue(self.celery, relay, name, backoff=backoff)
         elif options.type == 'custom':
             new_queue = self._load_from_custom(options, relay)
         else:
@@ -305,18 +358,6 @@ class SlimtaState(object):
         self.queues[name] = new_queue
         return new_queue
 
-    @property
-    def celery(self):
-        from .celery import get_celery_app
-        if not self._celery:
-            self._celery = get_celery_app(self.cfg)
-        return self._celery
-
-    def start_celery_queues(self):
-        for name, options in dict(self.cfg.queue).items():
-            if options.type == 'celery':
-                self._start_queue(name, options)
-
     def _start_edge(self, name, options=None):
         if name in self.edges:
             return self.edges[name]
@@ -329,8 +370,8 @@ class SlimtaState(object):
             from slimta.edge.smtp import SmtpEdge
             from .helpers import build_smtpedge_validators, build_smtpedge_auth
             from .helpers import fill_hostname_template
-            ip = options.listener.get('interface', '127.0.0.1')
-            port = int(options.listener.get('port', 25))
+            listener_defaults = {'interface': '127.0.0.1', 'port': 25}
+            listener = self._get_listener(options.listener, listener_defaults)
             kwargs = {}
             kwargs['tls'] = self._get_tls_options(options.get('tls'))
             kwargs['tls_immediately'] = options.get('tls_immediately', False)
@@ -338,9 +379,9 @@ class SlimtaState(object):
             kwargs['auth_class'] = build_smtpedge_auth(options)
             kwargs['command_timeout'] = 20.0
             kwargs['data_timeout'] = 30.0
-            kwargs['max_size'] = options.get('max_size', 10485760)
+            kwargs['max_size'] = int(options.get('max_size', 10485760))
             kwargs['hostname'] = fill_hostname_template(options.get('hostname'))
-            new_edge = SmtpEdge((ip, port), queue, **kwargs)
+            new_edge = SmtpEdge(listener, queue, **kwargs)
             new_edge.start()
         elif options.type == 'http':
             from slimta.edge.wsgi import WsgiEdge
@@ -350,10 +391,10 @@ class SlimtaState(object):
             uri_pattern = options.get('uri')
             validator_class = build_wsgiedge_validators(options)
             new_edge = WsgiEdge(queue, hostname, validator_class, uri_pattern)
-            ip = options.listener.get('interface', '127.0.0.1')
-            port = int(options.listener.get('port', 8025))
+            listener_defaults = {'interface': '127.0.0.1', 'port': 8025}
+            listener = self._get_listener(options.listener, listener_defaults)
             tls = self._get_tls_options(options.get('tls'))
-            server = new_edge.build_server((ip, port), tls=tls)
+            server = new_edge.build_server(listener, tls=tls)
             server.start()
         elif options.type == 'custom':
             new_edge = self._load_from_custom(options, queue)
@@ -362,7 +403,30 @@ class SlimtaState(object):
         self.edges[name] = new_edge
         return new_edge
 
+    def reload_config(self):
+        self.load_config()
+        old_edges = self.edges.copy()
+        old_queues = self.queues.copy()
+        old_relays = self.relays.copy()
+        self.edges = {}
+        self.queues = {}
+        self.relays = {}
+        self.start_everything()
+        for edge in old_edges.itervalues():
+            edge.kill()
+        for queue in old_queues.itervalues():
+            queue.kill()
+        for relay in old_relays.itervalues():
+            relay.kill()
+
+    def _handle_loop_interrupts(self, action):
+        if action == 'reload':
+            self.reload_config()
+
     def start_everything(self):
+        self.cached_listeners = self.listeners.copy()
+        self.listeners = {}
+
         if 'relay' in self.cfg:
             for name, options in dict(self.cfg.relay).items():
                 self._start_relay(name, options)
@@ -374,27 +438,9 @@ class SlimtaState(object):
             for name, options in dict(self.cfg.edge).items():
                 self._start_edge(name, options)
 
-    def worker_loop(self):
-        from .celery import get_celery_worker
-
-        self.start_celery_queues()
-
-        self.setup_logging()
-        self.redirect_streams()
-        self.daemonize()
-        with slimta.system.PidFile(self.args.pid_file):
-            sleep(0.1)
-            self.drop_privileges()
-
-            try:
-                with self._with_sighandlers():
-                    get_celery_worker(self.celery).run()
-            except (KeyboardInterrupt, SystemExit):
-                pass
+        self.cached_listeners = {}
 
     def loop(self):
-        from gevent.event import Event
-
         self.start_everything()
 
         self.setup_logging()
@@ -406,7 +452,10 @@ class SlimtaState(object):
 
             try:
                 with self._with_sighandlers():
-                    Event().wait()
+                    while True:
+                        action = self.loop_interrupt.get()
+                        self.loop_interrupt = AsyncResult()
+                        self._handle_loop_interrupts(action)
             except (KeyboardInterrupt, SystemExit):
                 pass
 
